/*
 * Copyright (c) 1993-2012 David Gay and Gustav Hllberg
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose, without fee, and without written agreement is hereby granted,
 * provided that the above copyright notice and the following two paragraphs
 * appear in all copies of this software.
 *
 * IN NO EVENT SHALL DAVID GAY OR GUSTAV HALLBERG BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF DAVID GAY OR
 * GUSTAV HALLBERG HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * DAVID GAY AND GUSTAV HALLBERG SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
 * "AS IS" BASIS, AND DAVID GAY AND GUSTAV HALLBERG HAVE NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifdef linux
  /* needed for REG_xxx constants for signal contexts */
  #define _GNU_SOURCE
#endif

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../context.h"
#include "../module.h"

#include "arith.h"
#include "basic.h"
#include "bigint.h"
#include "bitset.h"
#include "bool.h"
#include "debug.h"
#include "files.h"
#include "io.h"
#include "list.h"
#include "mudlle-float.h"
#include "mudlle-xml.h"
#include "pattern.h"
#include "runtime.h"
#include "mudlle-string.h"
#include "support.h"
#include "symbol.h"
#include "vector.h"


static ulong op_count;
static struct string *system_module;

static void system_set(ulong idx, value val)
{
  /* Technically this is not required, but you should think hard about whether
     adding a mutable or writable global is the right thing to do. */
  assert(readonlyp(val));
  if (!immutablep(val))
    {
      check_immutable(val);
      assert(immutablep(val));
    }

  assert(!GCONSTANT(idx));

  GVAR(idx) = val;
  module_vset(idx, var_module, system_module);
}

void system_string_define(struct string *name, value val)
{
  GCPRO1(val);
  ulong idx = mglobal_lookup(name);
  UNGCPRO();
  system_set(idx, val);
}

void system_write(struct string *name, value val)
/* Modifies: environment
   Requires: name not already exist in environment.
   Effects: Adds name to environment, with value val for the variable,
     as a 'define' of the system module.
*/
{
  GCPRO1(val);
  ulong aindex = mglobal_lookup(name);
  UNGCPRO();

  if (!module_vset(aindex, var_system_write, NULL))
    abort();
  GVAR(aindex) = val;
}

struct vector *define_mstring_vector(struct string *name,
                                     const char *const *vec,
                                     int count)
{
  if (name != NULL)
    assert(name->o.garbage_type == garbage_static_string);

  if (count < 0)
    for (count = 0; vec[count]; ++count)
      ;

  struct vector *v = alloc_vector(count);

  GCPRO1(v);
  for (int n = 0; n < count; ++n)
    SET_VECTOR(v, n, make_readonly(alloc_string(vec[n])));
  UNGCPRO();
  make_readonly(v);
  if (!check_immutable(&v->o))
    abort();
  if (name != NULL)
    system_string_define(name, v);
  return v;
}

void define_int_vector(struct string *name, const int *vec, int count)
{
  assert(name->o.garbage_type == garbage_static_string);

  struct vector *v = alloc_vector(count);
  GCPRO1(v);
  for (int n = 0; n < count; ++n)
    v->data[n] = makeint(vec[n]);
  UNGCPRO();
  if (!check_immutable(make_readonly(v)))
    abort();
  system_string_define(name, v);
}

#define tassert(what) do {                                      \
  if (!(what))                                                  \
    {                                                           \
      fprintf(stderr, "%s:%d: %s: Assertion `%s' failed.\n",    \
              op->filename, op->lineno, op->name, #what);       \
      abort();                                                  \
    }                                                           \
} while (0)

static void validate_typing(const struct primitive_ext *op)
{
  static const char allowed[] = "fnzZsvluktryxo123456789SdDbB";

  for (const char *const *type = op->type; *type; ++type)
    {
      const char *this = *type;
      int argc = 0;
      bool saw_period = false;
      for (const char *s = this; *s; ++s)
        {
          if (*s == '.')
            {
              saw_period = true;
              ++s;
              if (*s == 0)
                break;
            }
          if (*s == '[')
            {
              for (;;)
                {
                  ++s;
                  tassert(*s);
                  if (*s == ']')
                    break;
                  tassert(strchr(allowed, *s));
                }
            }
          else if (*s == '*')
            {
              tassert(s > this && s[1] == '.' && !saw_period);
              tassert(s[-1] != ']'); /* not supported by recurse_typing */
              continue;
            }
          else
            tassert(strchr(allowed, *s));
          if (saw_period)
            {
              tassert(!s[1]);
              break;
            }
          ++argc;
        }

      tassert(saw_period);

      if (op->nargs >= 0)
        tassert(op->nargs == argc);
    }
}

#undef tassert

static struct {
  size_t size, used;
  const struct primitive_ext **ops;
  bool locked;
} primitives;

void runtime_define(const struct primitive_ext *op)
{
  assert(op->nargs <= MAX_PRIMITIVE_ARGS);
  assert(!primitives.locked);

  ulong idx = global_lookup(op->name);

  validate_typing(op);

  struct primitive *prim = allocate_primitive(op);

  if (primitives.used == primitives.size)
    {
      primitives.size = primitives.size ? primitives.size * 2 : 16;
      primitives.ops = realloc(primitives.ops,
                               primitives.size * sizeof primitives.ops[0]);
    }
  primitives.ops[primitives.used++] = op;

  system_set(idx, prim);
}

static int cmp_ops(const void *_a, const void *_b)
{
  ulong a = (ulong)(*(const struct primitive_ext *const *)_a)->op;
  ulong b = (ulong)(*(const struct primitive_ext *const *)_b)->op;
  return a < b ? -1 : a > b;
}

static void sort_primitives(void)
{
  assert(!primitives.locked);
  primitives.ops = realloc(primitives.ops,
                           primitives.used * sizeof primitives.ops[0]);
  primitives.size = primitives.used;
  qsort(primitives.ops, primitives.used, sizeof primitives.ops[0],
        cmp_ops);
  primitives.locked = true;
}

const struct primitive_ext *lookup_primitive(ulong adr)
{
  assert(primitives.locked);
  struct primitive_ext key;
  key.op = (value (*)())adr;
  void *keyptr = &key;
  const struct primitive_ext **op = bsearch(
    &keyptr, primitives.ops, primitives.used,
    sizeof primitives.ops[0], cmp_ops);
  return op ? (*op) : NULL;
}

#ifdef MUDLLE_INTERRUPT
static bool interrupted = false;

void check_interrupt(void)
/* Effects: Causes a user_interrupt runtime error if interrupted is true
     (user caused SIGINT or SIGQUIT)
*/
{
  if (interrupted)
    {
      interrupted = false;
      runtime_error(error_user_interrupt);
    }
}

static void catchint(int sig)
{
  xcount = 1;
  interrupted = true;
}
#endif  /* MUDLLE_INTERRUPT */

#if defined i386 && !defined NOCOMPILER

#    undef USE_SYS_UCONTEXT
#    if defined __GLIBC__ && (defined REG_EIP || __GLIBC_MINOR__ >= 3)
#        include <sys/ucontext.h>
#        define USE_SYS_UCONTEXT
#    elif defined __MACH__
#        include <sys/ucontext.h>
#    elif defined SA_SIGINFO
#        include <asm/ucontext.h>
#    elif __GLIBC__ < 2 || (__GLIBC__ == 2 && __GLIBC_MINOR__ <= 1)
#        define sigcontext sigcontext_struct
#        include <asm/sigcontext.h>
#    elif __GLIBC__ != 2 && __GLIBC_MINOR__ != 2
#        include <sigcontext.h>
#    endif

#ifdef __MACH__
    typedef mcontext_t reg_context_t;
    #define GETREG(ctx, reg, REG) (*(ctx))->__ss.__ ## reg
    #define UCONTEXT_T ucontext_t
#elif defined USE_SYS_UCONTEXT
    typedef mcontext_t reg_context_t;
    #define GETREG(ctx, reg, REG) (ctx)->gregs[REG_ ## REG]
    #define UCONTEXT_T struct ucontext
#else
    typedef struct sigcontext reg_context_t;
    #define GETREG(ctx, reg, REG) (ctx)->reg
    #define UCONTEXT_T struct ucontext
#endif

#include <stddef.h>
#include "../builtins.h"

#ifdef SA_SIGINFO
static struct sigaction oldsegact;
#ifdef __MACH__
static struct sigaction oldbusact;
#endif
#endif

void got_real_segv(int sig);
void got_real_segv(int sig)
{
  /* reinstall the default handler; this will cause a real crash */
#ifdef SA_SIGINFO
  if (sig == SIGSEGV)
    {
      sigaction(sig, &oldsegact, NULL);
      return;
    }
#ifdef __MACH__
  if (sig == SIGBUS)
    {
      sigaction(sig, &oldbusact, NULL);
      return;
    }
#endif
  abort();
#else
  signal(sig, SIG_DFL);
#endif
}

static void check_segv(int sig, reg_context_t *scp)
{
  const ubyte *eip = (const ubyte *)GETREG(scp, eip, EIP);

  CASSERT_EXPR(offsetof (struct obj, size) == 0);

  if ((eip[0] == 0x80           /* cmpb $imm,object_type(reg) */
       && (eip[1] & 0xf8) == 0x78
       && eip[2] == 5)          /* offsetof(struct obj, type) */
      || ((eip[0] == 0x81        /* cmpl $imm,object_size(reg) */
           || eip[0] == 0x83)    /* cmpl $imm8,object_size(reg) */
          && (eip[1] & 0xf8) == 0x38)
      || (eip[0] == 0x0f        /* movzbl object_type(%ecx),%eax */
          && eip[1] == 0xb6
          && (eip[2] & 0xc0) == 0x40
          && eip[3] == 5)       /* offsetof(struct obj, type) */
      || (eip[0] == 0x8b        /* mov (reg0),reg1 for size */
          && (eip[1] & 0xc0) == 0)
      || (eip[0] == 0x8b        /* mov car/cdr(reg0),reg1 */
          && (eip[1] & 0xc0) == 0x40
          && (eip[2] == offsetof(struct list, car)
              || eip[2] == offsetof(struct list, cdr))))
    {
      /* mudlle code */
      if (eip >= gcblock && eip < gcblock + gcblocksize)
	{
	  ccontext.frame_end_sp = (ulong *)GETREG(scp, esp, ESP);
	  ccontext.frame_end_bp = (ulong *)GETREG(scp, ebp, EBP);
          /* make stack trace printing start from the right place */
	  ccontext.frame_end_sp[-1] = (ulong)eip;
	  runtime_error(error_bad_type);
	}
      /* bref and bset */
      if (eip >= (ubyte *)bref && eip < (ubyte *)bwglobal)
	{
	  ccontext.frame_end_sp = (ulong *)GETREG(scp, esp, ESP) + 1;
	  ccontext.frame_end_bp = (ulong *)GETREG(scp, ebp, EBP);
          bool is_set = eip >= (ubyte *)bset;
          if (is_set)
            set_runtime_error(error_bad_type,
                              (value)GETREG(scp, eax, EAX),
                              (value)GETREG(scp, ecx, ECX),
                              (value)GETREG(scp, edx, EDX));
          else
            ref_runtime_error(error_bad_type,
                              (value)GETREG(scp, eax, EAX),
                              (value)GETREG(scp, ecx, ECX));
	}
      /* bcall */
      if (eip >= (ubyte *)bcall && eip < (ubyte *)bcall_secure)
	{
	  ccontext.frame_end_sp = (ulong *)GETREG(scp, esp, ESP) + 1;
	  ccontext.frame_end_bp = (ulong *)GETREG(scp, ebp, EBP);
	  runtime_error(error_bad_function);
	}
    }

  got_real_segv(sig);
}

#ifdef SA_SIGINFO
static void catchsegv(int sig, siginfo_t *siginfo, void *_sigcontext)
{
  UCONTEXT_T *sigcontext = _sigcontext;
  check_segv(sig, &sigcontext->uc_mcontext);
}
#else
static void catchsegv(int sig, struct sigcontext scp)
{
  check_segv(sig, &scp);
}
#endif

#endif  /* i386 && !NOCOMPILER */

void flag_check_failed(const struct primitive_ext *op, const char *name)
{
  abort();
}

void runtime_init(void)
{
  op_count = 0;
  system_module = alloc_string("system");
  staticpro(&system_module);

#ifdef MUDLLE_INTERRUPT
  signal(SIGINT, catchint);
#ifdef SIGQUIT
  signal(SIGQUIT, catchint);
#endif
#endif

#if defined(i386) && !defined(NOCOMPILER)
#ifdef SA_SIGINFO
  {
    static char my_signal_stack[MINSIGSTKSZ + 8 * 1024];
    stack_t my_stack = {
      .ss_sp    = my_signal_stack,
      .ss_flags = 0,
      .ss_size  = sizeof my_signal_stack,
    };
    bool noaltstack = sigaltstack(&my_stack, NULL) < 0;
    if (noaltstack)
      {
	perror("sigaltstack()");
      }

    struct sigaction sact = {
      .sa_sigaction = catchsegv,
      .sa_flags     = (SA_SIGINFO | SA_RESTART | SA_NODEFER
                       | (noaltstack ? 0 : SA_ONSTACK)),
    };
    sigemptyset(&sact.sa_mask);
    sigaction(SIGSEGV, &sact, &oldsegact);
#ifdef __MACH__
    sigaction(SIGBUS, &sact, &oldbusact);
#endif
  }
#else
  signal(SIGSEGV, (void (*)(int))catchsegv);
#endif
#endif

  basic_init();
  debug_init();
  arith_init();
  bool_init();
  io_init();
  symbol_init();
  string_init();
  list_init();
  vector_init();
  support_init();
  bitset_init();
  files_init();
  float_init();
  bigint_init();
  pattern_init();
  mudlle_consts_init();
  xml_init();
  module_set("system", module_protected, 0);

  sort_primitives();
}
